import re

import click


def validate_account_id(ctx, param, value):
    if not re.match('\\d{3}-?\\d{3}-?\\d{4}$', value):
        raise click.BadParameter('account id must be a 10-digit number or ???-???-????')

    return int(value.replace('-', ''))



import sys

import google.ads.google_ads.client
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.filters import is_searching

from gaql import DEFAULT_LIMIT, HISTORY_FILE
from gaql.lib.google_clients.completion.google_completer import GoogleCompleter
from gaql.lib.lexer import GaqlLexer
from gaql.lib.output import write_rows, print_gaql

_SESSION = None


def setup_session(completion):
    global _SESSION
    bindings = KeyBindings()
    lexer = PygmentsLexer(GaqlLexer)

    @bindings.add('c-m', filter=~is_searching)
    def _(event):
        if ';' in event.current_buffer.text:
            event.current_buffer.validate_and_handle()
        else:
            event.current_buffer.insert_text('\n')

    _SESSION = PromptSession(
        multiline=True,
        key_bindings=bindings,
        lexer=lexer,
        history=FileHistory(HISTORY_FILE),
        completer=GoogleCompleter() if completion else None,
    )


def read_query():
    query = _SESSION.prompt('> ')
    return parse_query(query)


def parse_query(query, unlimited = False):
    base_query = query.replace(';', '')
    if unlimited or 'limit' in base_query.lower():
        return base_query
    else:
        return base_query + ' LIMIT ' + DEFAULT_LIMIT


def run_as_repl(run_query, state, completion):
    setup_session(completion)
    print('running in interactive mode, write some GAQL...')

    try:
        while True:
            query = read_query()
            if query:
                print_gaql(query, _SESSION.lexer.pygments_lexer)
                try:
                    rows = run_query(query)
                    write_rows(rows, state)
                except google.ads.google_ads.errors.GoogleAdsException as ex:
                    handle_google_ads_error(ex)
    except (EOFError, KeyboardInterrupt):
        print('\nBye!')


def handle_google_ads_error(ex):
    print(
        f'Request with ID "{ex.request_id}" failed with status'
        + f'{ex.error.code().name} and errors:'
    )
    for error in ex.failure.errors:
        print(f'\tError with message {error.message}')



import click

from gaql.lib.click_decorators.state import common_options, pass_state
from gaql.lib.google_clients.config import setup_client
from gaql.lib.google_clients.queries import google_fields_query
from gaql.lib.output import write_rows, print_gaql
from gaql.lib.repl import run_as_repl, parse_query
from gaql.tools.queries import clients_cmd


@click.group(context_settings=dict(max_content_width=120))
@common_options
@pass_state
def tools(state):
    pass


@tools.group()
@pass_state
def queries(state):
    """Saved common queries for GoogleAds"""
    state.finalize([])


@tools.command()
@click.argument('entity', required=True)
@pass_state
def fields_for(state, entity):
    """Query selectable fields for a specific entity"""
    state.finalize(entity)

    client = setup_client()
    query_method = google_fields_query(client)

    query = f"SELECT name, attribute_resources, selectable_with WHERE name = '{entity}'"

    print_gaql(query)
    rows = query_method(parse_query(query))
    write_rows(rows, state)


@tools.command()
@click.argument('query', nargs=-1)
@pass_state
def fields(state, query):
    """Make queries about field metadata for GoogleAds entities"""
    state.finalize(query)

    client = setup_client()
    query_method = google_fields_query(client)

    if len(query) == 0:
        run_as_repl(query_method, state, completion=False)
    else:
        rows = query_method(parse_query(' '.join(query)))
        write_rows(rows, state)


@queries.command()
@pass_state
def clients(state):
    """Returns all the clients"""
    clients_cmd(state)



import click
import json
import os
import sys

from gaql import CONFIG_DIR, CREDENTIAL_FILE


def configure():
    login_customer_id = input('login_customer_id: ').replace('-', '')

    secrets = {
        'developer_token': input('developer_token: '),
        'client_id': input('client_id: '),
        'client_secret': input('client_secret: '),
        'login_customer_id': login_customer_id,
        'refresh_token': input('refresh_token: '),
    }

    os.makedirs(CONFIG_DIR, exist_ok=True)

    with CREDENTIAL_FILE.open('w') as f:
        print(f'Wrote credentials to {CREDENTIAL_FILE}')
        f.write(json.dumps(secrets, indent=4))


def read_credentials_from_file():
    def read_credential_file_json():
        with CREDENTIAL_FILE.open('r') as f:
            return json.load(f)

    try:
        return read_credential_file_json()
    except FileNotFoundError:
        print(f"Configuration file not found at : {CREDENTIAL_FILE}", file=sys.stderr)
        should_configure = click.confirm('Configure it now?')
        if should_configure:
            configure()
            return read_credential_file_json()
        ERROR_MSG = (
            "Cannot continue without credentials. Please add credentials to your environment, home directory,"
            " or fill out the credentials prompt."
        )
        print(ERROR_MSG, file=sys.stderr)
        sys.exit(1)



def load_credentials():
    """Loads credentials with the following ordering: from env, Google's yaml file, a custom .json file in .config/gaql"""
    from google.ads.google_ads import config
    try:
        return config.load_from_env()
    except:
        try:
            return config.load_from_yaml_file()
        except:
            if not CREDENTIAL_FILE.exists():
                print(f"Couldn't load credentials from the environment, or from ~/google-ads.yaml. Trying {CREDENTIAL_FILE}")
            return config.load_from_dict(read_credentials_from_file())


def get_root_client():
    return load_credentials()['login_customer_id']


def setup_client():
    from google.ads.google_ads.client import GoogleAdsClient
    credentials = load_credentials()
    return GoogleAdsClient.load_from_dict(credentials)


def setup_ads_service(client):
    return client.get_service('GoogleAdsService', version='v3')


def setup_fields_service(client):
    return client.get_service('GoogleAdsFieldService', version='v3')



def chain(*fns):
    """Chain multiple single argument function calls together"""

    def chained_fns(arg):
        result = arg
        for f in fns:
            result = f(result)
        return result

    return chained_fns



"""Output GoogleAdsRows in different formats"""
import collections
import inspect
from json import dump

from google.protobuf.json_format import MessageToJson, MessageToDict
from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter

from gaql.lib.functional import chain
from gaql.lib.lexer import GaqlLexer


def flatten(d, parent_key=''):
    """recursively flatten a nested dictionary of reasonable depth"""
    items = []
    for k, v in d.items():
        key = parent_key + '.' + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, key).items())
        else:
            items.append((key, v))
    return dict(items)


def print_gaql(query, lexer=GaqlLexer()):
    print(highlight(inspect.cleandoc(query), lexer, Terminal256Formatter()))


def write_csv(stream, rows):
    import csv
    import json

    convert_row = chain(MessageToJson, json.loads, flatten)

    rows = iter(rows)
    first_row = next(rows, None)

    if first_row:
        first_row = convert_row(first_row)
        writer = csv.DictWriter(stream, first_row.keys())
        writer.writeheader()
        writer.writerow(first_row)
        for row in rows:
            row = convert_row(row)
            writer.writerow(row)

def to_dict(row):
    return MessageToDict(row, preserving_proto_field_name=True)


def write_json(stream, rows):
    for row in rows:
        print(MessageToJson(row), file=stream)


def write_json_lines(stream, rows):
    for row in rows:
        dump(to_dict(row), fp=stream, separators=(',', ':'))
        stream.write('\n')


def write_proto(stream, rows):
    for row in rows:
        print(row, file=stream)


def write_rows(rows, state):
    if state.format == 'json':
        write_json(state.output, rows)
    elif state.format == 'jsonl':
        write_json_lines(state.output, rows)
    elif state.format == 'csv':
        write_csv(state.output, rows)
    else:
        write_proto(state.output, rows)
